/**
 * Copyright (c) 2004 Logos GmbH Rostock (www.logos-ib.de).
 * All rights reserved.
 *
 * THIS SOFTWARE IS  PROVIDED  "AS IS"  AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF  MERCHANTABILITY  AND  FITNESS  FOR  A PARTICULAR  PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL  LOGOS GMBH  OR ITS CONTRIBUTORS BE
 * LIABLE FOR ANY  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES  (INCL., BUT NOT LIMITED TO, PROCUREMENT
 * OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE, DATA, OR  PROFITS;
 * OR BUSINESS INTERRUPTION)  HOWEVER  CAUSED  AND  ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,  OR TORT (INCL.
 * NEGLIGENCE OR OTHERWISE)  ARISING  IN  ANY  WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package de.bsvrz.dua.bastband.bastband.dvs.funclib;

/**
 * History:
 *
 * $Log: DataSender.java,v $
 * Revision 1.2  2007/07/25 13:02:05  Trier
 * <Kein Kommentar eingegeben>
 *
 * Revision 1.1.1.1  2007/06/20 09:56:26  Trier
 * dvs alles
 *
 * Revision 1.6  2007/06/14 13:57:46  Trier
 * <Kein Kommentar eingegeben>
 *
 * Revision 1.5  2006/11/13 09:24:10  Trier
 * <Kein Kommentar eingegeben>
 *
 * Revision 1.4  2006/06/30 09:42:42  Trier
 * <Kein Kommentar eingegeben>
 *
 * Revision 1.3  2006/06/02 14:23:55  Trier
 * <Kein Kommentar eingegeben>
 *
 * Revision 1.2  2006/06/02 13:10:54  Strey
 * <Kein Kommentar eingegeben>
 *
 * Revision 1.1  2006/06/02 13:04:21  Trier
 * initial
 *
 *
 */
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Timer;
import java.util.TimerTask;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.OneSubscriptionPerSendData;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.SendSubscriptionNotConfirmed;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.dataIdentificationSettings.DataIdentification;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * <p>
 * Klasse zum Versenden von Daten.
 * </p>
 *
 * <p>
 * Die Methode {@link #sendData(ResultData rs) sendData(ResultData) } dient dem
 * Versenden von Datens&auml;tzen. Die Klasse verwaltet eine Liste von
 * DataIdentifications f&uuml;r die Sendeanmeldungen ({@link SenderSubscribtion}
 * ) bestehen. Sendeanmeldungen, die lange nicht benutzt wurden, werden
 * zur&uuml;ckgenommen (unsubscribed) und die entsprechene DataIdentification
 * aus der Liste gel&ouml;scht.
 *
 * <p>
 * Ist ein Datensatz zu versenden ({@link #sendData(ResultData rs)
 * sendData(ResultData)}), wird
 * <ul>
 * <li>die zum Senden anzumeldene DatenIdentifikation aus dem ResultData
 * bestimmt.</li>
 * <li>Es wird gepr&uuml;ft, ob f&uuml;r die DatenIdentifikation eine
 * {@link SenderSubscribtion} existiert und falls dies nicht der Fall ist, wird
 * eine {@link SenderSubscribtion} erzeugt.
 * <li>Der Datensatz wird durch Aufruf
 * {@link SenderSubscribtion#sendData(ResultData rs)} versendet.</li>
 * <li>Alle 10sec wird gepr&uuml;ft, ob die HoldSubscriptionTime (msec)
 * &uuml;berschritten wurde und ggf. die DataIdentification zum Versenden
 * abgemeldet.</li>
 * </ul>
 *
 * @author Ralf Trier
 * @version 1.0
 * @see SenderSubscribtion
 */
public class DataSender {

	private static final Debug LOGGER = Debug.getLogger();

	/**
	 * zu verwendenes ClientDavInterface.
	 */
	protected ClientDavInterface _DavConn;
	/**
	 * Map zum Speichern DataIdentification =&gt; SenderSubscribtion.
	 */
	protected HashMap<DataIdentification, SenderSubscribtion> _MapSubscribtions = new HashMap<>();
	/**
	 * Zeit seit der Nutzung in msec nach der, die Subscription gel&ouml;scht werden
	 * soll.
	 */
	protected long _HoldSubscriptionTime;
	/**
	 * Zeit in msec, die auf eine positive Sendesteuerung gewartet wird, bevor davon
	 * ausgegangen wird, dass es weder Senke noch Empf&auml;nger gibt.
	 */
	protected int _TimeOut = 5000;

	/**
	 * Ist _IgnoreStopSending==true, dann wird auch die TimeoutZeit auf eine
	 * positive Sendesteuerung gewartet, wenn ein STOP_SENDING ausgel&ouml;st wurde.
	 * Normalerweise wird davon ausgegangen, dass wenn STOP_SENDING ausgel&ouml;st
	 * wurde, dass weder Quelle noch Senke vorhanden ist.
	 */
	protected boolean _IgnoreStopSending = false;

	/**
	 * Timer, der regelm&auml;&szlig;ig Anmeldungen zur&uuml;cknimmt, wenn die
	 * HoldSubscriptionTime &uuml;berschritten wurde.
	 */
	protected Timer _Timer;

	/**
	 * erzeugt einen DataSender
	 *
	 * @param davConn
	 *            zu verwendenes ClientDavInterface
	 * @param holdSubscriptionTime
	 *            Zeit in msec bei deren &Uuml;berschreitung die Anmeldung
	 *            gel&ouml;scht wird (0 Anmeldungen werden nie gel&ouml;scht)
	 */
	public DataSender(final ClientDavInterface davConn, final long holdSubscriptionTime) {
		_HoldSubscriptionTime = holdSubscriptionTime;
		_DavConn = davConn;
		if (holdSubscriptionTime > 0) {
			final TimerTask timerTask = new CleanOldSubscribtionTimerTask(this);
			final Timer timer = new Timer("CleanOldSubscribtionTimer");
			timer.schedule(timerTask, 10000, 10000);
		}
	}

	/**
	 * erzeugt einen DataSender
	 *
	 * @param davConn
	 *            zu verwendenes ClientDavInterface
	 * @param holdSubscriptionTime
	 *            Zeit in msec bei deren &Uuml;berschreitung die Anmeldung
	 *            gel&ouml;scht wird (0 Anmeldungen werden nie gel&ouml;scht)
	 * @param timeout
	 *            Zeit in msec, die auf eine positive Sendesteuerung gewartet wird,
	 *            bevor davon ausgegangen wird, dass es weder Senke noch
	 *            Empf&auml;nger gibt.
	 * @param ignoreStopSending
	 *            Meldung zum Stoppen des Datenversands ignorieren
	 */
	public DataSender(final ClientDavInterface davConn, final long holdSubscriptionTime, final int timeout,
			final boolean ignoreStopSending) {
		_HoldSubscriptionTime = holdSubscriptionTime;
		_DavConn = davConn;
		_IgnoreStopSending = ignoreStopSending;
		_TimeOut = timeout;
		if (holdSubscriptionTime > 0) {
			final TimerTask timerTask = new CleanOldSubscribtionTimerTask(this);
			final Timer timer = new Timer("CleanOldSubscribtionTimer");
			timer.schedule(timerTask, 10000, 10000);
		}
	}

	/**
	 * versendet den Datensatz.
	 *
	 * @param rs
	 *            zu versendender Datensatz
	 * @throws SendSubscriptionNotConfirmed
	 *             wird geworfen, wenn innerhalb der TimeoutZeit keine positive
	 *             Sendesteuerung vom Datenverteiler für die zu versendenden Daten
	 *             vorliegt oder kein Empf&auml;nger vorhanden ist.
	 * @throws OneSubscriptionPerSendData
	 *             wird geworfen, wenn bereits eine lokale Sendeanmeldung für die
	 *             gleichen Daten von einem anderen Anwendungsobjekt vorliegt.
	 */
	public void sendData(final ResultData rs) throws SendSubscriptionNotConfirmed, OneSubscriptionPerSendData {

		final SystemObject so = rs.getObject();
		final DataDescription dataDescr = rs.getDataDescription();
		final DataIdentification dataIdentification = new DataIdentification(so, dataDescr);

		SenderSubscribtion subscription = null;
		synchronized (_MapSubscribtions) {
			subscription = _MapSubscribtions.get(dataIdentification);
			if (subscription == null) {
				subscription = new SenderSubscribtion(_DavConn, dataIdentification, _TimeOut, _IgnoreStopSending);
				_MapSubscribtions.put(dataIdentification, subscription);
			}
		}
		subscription.sendData(rs);
	}

	/**
	 * l&ouml;scht eine Sendeanmeldung und gibt true zur&uuml;ck, wenn dies
	 * erfolgreich war.
	 *
	 * @param dataIdentification
	 *            DataDescription
	 * @return boolean
	 */
	public synchronized boolean deleteSubscription(final DataIdentification dataIdentification) {
		synchronized (_MapSubscribtions) {
			final SenderSubscribtion subscribtion = _MapSubscribtions.remove(dataIdentification);
			if (subscribtion == null) {
				return false;
				// throw new
				// NullPointerException("Anmeldung für "+dataIdentification+"
				// existiert nicht");
			}
			try {
				subscribtion.unsubscribeSender();
				return true;
			} catch (final Exception e) {
				DataSender.LOGGER.error("", e);
				return false;
			}
		}
	}

	/**
	 * l&ouml;scht alle Anmeldungen und stoppt wenn n&ouml;tig den Timer.
	 */
	public synchronized void close() {
		DataSender.LOGGER.finest("close called");
		synchronized (_MapSubscribtions) {
			final Iterator<Map.Entry<DataIdentification, SenderSubscribtion>> it = _MapSubscribtions.entrySet()
					.iterator();
			while (it.hasNext()) {
				final Map.Entry<DataIdentification, SenderSubscribtion> entry = it.next();
				final SenderSubscribtion subscribtion = entry.getValue();
				subscribtion.unsubscribeSender();
				it.remove();
				DataSender.LOGGER.fine("removed", subscribtion);
			}
		}
		if (_Timer != null) {
			_Timer.cancel();
		}
	}

	synchronized void deleteOldSubscription() {
		final long lCurrentTime = System.currentTimeMillis();
		DataSender.LOGGER.finest("deleteOldSubscription called");
		synchronized (_MapSubscribtions) {
			// Set<Map.Entry<DataIdentification, SenderSubscribtion>> set =
			// _MapSubscribtions.entrySet();
			final Iterator<Map.Entry<DataIdentification, SenderSubscribtion>> it = _MapSubscribtions.entrySet()
					.iterator();
			while (it.hasNext()) {
				final Map.Entry<DataIdentification, SenderSubscribtion> entry = it.next();
				final SenderSubscribtion subscribtion = entry.getValue();
				if (lCurrentTime > (subscribtion.getLastUse() + _HoldSubscriptionTime)) {
					subscribtion.unsubscribeSender();
					it.remove();
					DataSender.LOGGER.fine("removed", subscribtion);
				}
			}
		}
	}

	/**
	 * Task zum L&ouml;schen unbenutzer SenderSubscription
	 *
	 * @author unbekannt
	 * @version 1.0
	 */
	class CleanOldSubscribtionTimerTask extends TimerTask {

		DataSender _DataSender;

		CleanOldSubscribtionTimerTask(final DataSender dataSender) {
			_DataSender = dataSender;
		}

		@Override
		public void run() {
			_DataSender.deleteOldSubscription();
		}

	}

}
